* { 
font-family: arial, verdana, geneva, helvetica, sans-serif;
font-size: 12px;
}

body {
background-color: #ffffff;
color: #000000;
}

h1 {
font-size: 18px;  
color: #506070;
clear: both;
}

h2 {
font-size: 16px;  
color: #506070;
clear: both;
}

h3 {
font-size: 14px;  
color: #506070;
clear: both;
}

h4 {
font-size: 13px;  
color: #506070;
clear: both;
}

h5 {
font-size: 12px;  
color: #405060;
clear: both;
}

pre {
    padding: 10px;
    background-color: #f0f0f0;
    border: 1px solid #ccc;
    font-family: "Courier New", Courier, monospace;
}

a:link {
color: #445566;
}

a:visited {
color: #000050;
}

a:active {
color: #a000ff;
}

a:hover {
color: #000;
}


